from typing import Literal

from .. import OutputFormat

MimeType = Literal["text/csv", "application/json", "text/plain", "text/xml", "application/vnd.ms-excel"]


def get_mimetype(output_type: OutputFormat) -> MimeType:
    if output_type == "csv":
        return "text/csv"
    elif output_type == "json":
        return "application/json"
    elif output_type == "txt":
        return "text/plain"
    elif output_type == "xml":
        return "text/xml"
    elif output_type == "excel":
        return "application/vnd.ms-excel"
    else:
        raise ValueError(f"Unknown output type: {output_type}")



import json
from typing import List

from ..models import Pivot, TransformerOutput
from ..transformers.transformer import Transformer


class JSONTransformer(Transformer):
    def __init__(self):
        super(JSONTransformer, self).__init__()
        self.output_type = "json"
        self.output = TransformerOutput(data=None, output=self.output_type,
                                        filename=f'{self.name}_output.{self.output_type}')

    def transform(self, pivot_list: List[Pivot]) -> TransformerOutput:
        json_ver = json.dumps({i: article.model_dump() for i, article in enumerate(pivot_list)}, ensure_ascii=False, indent=4)
        self.output.data = json_ver
        return self.output




from typing import Literal, Set, Union, Optional

from pydantic import BaseModel, field_serializer


class FileToTransform(BaseModel):
    name: str
    file: str


class Error(BaseModel):
    message: str
    article: str
    transformer: str


class Pivot(BaseModel):
    journal: str
    journal_clean: str
    titre: str
    complement: str
    annee: int
    mois: int
    jour: int
    heure: int
    minute: int
    seconde: int
    date: str
    epoch: int
    auteur: str
    texte: str
    keywords: Set[str]
    langue: str
    url: str

    def __hash__(self):
        return hash((self.journal, self.date, self.titre))

    @field_serializer('keywords')
    def serialize_keywords(self, kw: Set[str]):
        return ', '.join(kw)


OutputFormat = Literal["csv", "excel", "json", "txt", "xml", "zip"]
Output = Literal["json", "txm", "iramuteq", "gephi", "csv", "excel", "stats", "processed_stats", "plots", "markdown"]


class TransformerOutput(BaseModel):
    data: Union[str, bytes, None]
    output: OutputFormat
    filename: str


class Params:
    def __init__(
            self,
            filter_keywords: bool = False,
            filter_lang: bool = False,
            minimal_support: int = 1,
            minimal_support_kw: Optional[int] = None,
            minimal_support_journals: Optional[int] = None,
            minimal_support_authors: Optional[int] = None,
            minimal_support_dates: Optional[int] = None,
    ):
        assert all((isinstance(x, int) and x > 0) or x is None
                   for x in [minimal_support, minimal_support_kw, minimal_support_journals, minimal_support_authors,
                             minimal_support_dates])
        assert all(isinstance(x, bool) for x in [filter_keywords, filter_lang])

        self.filter_keywords: bool = filter_keywords
        self.filter_lang: bool = filter_lang
        self.minimal_support: int = minimal_support
        self.minimal_support_kw: int = minimal_support_kw or minimal_support
        self.minimal_support_journals: int = minimal_support_journals or minimal_support
        self.minimal_support_authors: int = minimal_support_authors or minimal_support
        self.minimal_support_dates: int = minimal_support_dates or minimal_support



from __future__ import annotations

import concurrent.futures

from tqdm.auto import tqdm

from .models import Output, FileToTransform, TransformerOutput, Params
from .transformers.to_csv import CSVTransformer
from .transformers.to_excel import ExcelTransformer
from .transformers.to_iramuteq import IramuteqTransformer
from .transformers.to_json import JSONTransformer
from .transformers.to_markdown import MarkdownTransformer
from .pivot import PivotTransformer
from .transformers.to_txm import TXMTransformer
from .transformers.to_stats import StatsTransformer

transformer_factory = {
    "json": JSONTransformer().transform,
    "txm": TXMTransformer().transform,
    "iramuteq": IramuteqTransformer().transform,
    "gephi": None,
    "csv": CSVTransformer().transform,
    "excel": ExcelTransformer().transform,
    "stats": "get_stats",
    "processed_stats": "get_processed_stats",
    "plots": "get_plots",
    "markdown": MarkdownTransformer().transform
}

stats_outputs = {"stats", "processed_stats", "plots", "markdown"}


def pipeline(files: list[FileToTransform], outputs: list[Output], params: Params) -> list[TransformerOutput]:
    """
    main function that transforms the files into pivots and then in differents required ouptputs
    """

    if not isinstance(files, list):
        files = [files]

    for file in files:
        if not isinstance(file, FileToTransform):
            raise ValueError(f"Unknown file type: {file}")

    if not isinstance(outputs, list):
        outputs = [outputs]

    for output in outputs:
        if output not in transformer_factory:
            raise ValueError(f"Unknown output type: {output}")

        # if not isinstance(outputs, (Output, str, OutputFormat)):
        #     raise ValueError(f"Unknown output type: {output}")

    if not isinstance(params, Params):
        raise ValueError(f"Unknown params type: {params}")

    transformer = PivotTransformer(params)
    pivots = transformer.transform(files_to_transform=files)

    to_process = []
    st = None
    if stats_outputs.intersection(outputs):
        st = StatsTransformer()
        st.transform(pivots)

    for output in outputs:
        if output in stats_outputs and output != "markdown":
            func = getattr(st, transformer_factory[output])
            to_process.append((func, []))

        else:
            func = transformer_factory[output]
            args = [pivots]
            to_process.append((func, args))

    results: list[TransformerOutput] = []

    with concurrent.futures.ProcessPoolExecutor() as executor:
        futures = [executor.submit(func, *args) for func, args in to_process]
        for future in tqdm(concurrent.futures.as_completed(futures), total=len(futures)):
            res = future.result()
            results.append(res)

    return results



import sys

# if sys.version_info < (3, 9):
#     from __future__ import annotations

import re
from europarser.utils import STOP_WORDS


def filter_KW(liste):
    out = set()
    liste = set(liste).difference(STOP_WORDS)
    for x in liste:
        numbers = re.findall("[0-9]", x)
        if x.lower() in STOP_WORDS:
            continue
        elif len(x) < 2:
            continue
        elif len(re.findall("[A-Za-z]", x)) == 0:
            continue
        elif len(numbers) > 0:
            if len(numbers[0]) == len(x):
                continue
        out.add(x.lower().replace(',', '').strip())
    return out


def tokenize(chaine):
    chaine = re.sub(r"\.|'|\?|»|«|\"", " ", chaine)
    chaine = re.sub(" {2,}", " ", chaine)
    mots = chaine.split()
    return mots


def is_subsequence(needle, haystack):
    return any(haystack[i:i + len(needle)] == needle
               for i in range(len(haystack) - len(needle) + 1))


def get_ngrams(inter, word_titre, word_body):
    selected = [x for x in inter]
    haystack = [x.lower() for x in word_body]
    for i, word in enumerate(word_titre):
        if word in inter or word.lower() in inter:
            n_gram = [word]
            last = i
            for j, w in enumerate(word_titre[i + 1:]):
                if w in inter:
                    last = j
                n_gram.append(w)
            if len(n_gram) > 1 and len(n_gram) < 5 and last != i:
                needle = [x.lower() for x in n_gram]
                if is_subsequence(needle, haystack):
                    selected.append("_".join(n_gram[:last + 2]))
    ## TODO: régler le problème des n-grammes
    ##  if "gilets" in haystack and "jaunes in haystack":
    ##    if "gilets_jaunes" not in selected:
    ##      print(selected)
    ##      x = haystack.index("gilets")
    ##      print(haystack[x:x+10])
    ##      print(word_titre[word_titre.index("gilets"):])
    return selected


def get_KW(titre, text):
    body = text
    chapeau = ""
    if len(titre) < 100:
        chapeau = text[:200]
        body = text[200:]
    word_titre = set(tokenize(titre) + tokenize(chapeau))
    word_body = set(tokenize(body))
    inter = word_titre.intersection(word_body)
    inter = filter_KW(inter)
    inter = get_ngrams(inter, tokenize(titre), tokenize(body))
    return inter



from typing import List

from ..models import Pivot, TransformerOutput
from ..transformers.transformer import Transformer

import pandas as pd


class CSVTransformer(Transformer):
    def __init__(self):
        super(CSVTransformer, self).__init__()
        self.output_type = "csv"
        self.output = TransformerOutput(data=None, output=self.output_type, filename=f'{self.name}_output.{self.output_type}')

    def transform(self, pivot_list: List[Pivot]) -> TransformerOutput:
        df = pd.DataFrame.from_records([p.model_dump() for p in pivot_list])
        self.output.data = df.to_csv(sep=",", index=False)
        return self.output
            



import hashlib
import json
import sys

# if sys.version_info < (3, 9):
#     from __future__ import annotations

import re
from pathlib import Path
from typing import Tuple, Optional
from datetime import date, datetime
from dateutil import tz as tzutil

from .models import Pivot

date_regex = [
    find_french_date_1 := re.compile(
        r"(?:lundi|mardi|mercredi|jeudi|vendredi|samedi|dimanche) [0-9]+ (?:janvier|février|mars|avril|mai|juin|juillet|août|septembre|octobre|novembre|décembre) [0-9]{4}"),
    find_french_date_2 := re.compile(r'\d{1,2}\s\w+\s\d{4}'),
    find_english_date_1 := re.compile(r'\w+,\s?\w+\s\d{1,2},\s?\d{4}'),
    find_english_date_2 := re.compile(r'\w+\s\d{1,2},\s?\d{4}')
]
dic_months = {"janvier": "01", "février": "02", "mars": "03", "avril": "04", "mai": "05", "juin": "06", "juillet": "07",
              "août": "08", "septembre": "09", "octobre": "10", "novembre": "11", "décembre": "12"}
trad_months = {"January": "janvier", "February": "février", "March": "mars", "April": "avril", "May": "mai",
               "June": "juin", "July": "juillet", "August": "août", "September": "septembre", "October": "octobre",
               "November": "novembre", "December": "décembre"}
time_regex = re.compile(r'\d{1,2}:\d{1,2}(?::\d{1,2})?(?:\s?[ap]\.?m\.?)?(?:\s?[GU][MT][TC]\s[+-]\d{1,2})?')


def find_date(txt: [str]) -> Optional[date]:
    """
    Utility function to extract a date from aa given string
    :return: a date object with the date
    """
    day = month = year = ""
    index = 0
    final_month = None
    while index < len(date_regex) and final_month not in dic_months:
        match = date_regex[index].search(txt)
        if match:
            match = match[0]

            if date_regex[index] is find_french_date_1:
                _, day, month, year = match.split()

            elif date_regex[index] is find_french_date_2:
                day, month, year = match.split()

            elif date_regex[index] is find_english_date_1:
                _, month, year = match.split(',')
                month, day = month.strip().split()
                month = trad_months[month]

            elif date_regex[index] is find_english_date_2:
                month_day, year = match.split(',')
                month, day = month_day.split(' ')
                month = trad_months[month]

            day, month, year = [x.strip() for x in [day, month, year]]
            final_month = month
            index += 1
        else:
            index += 1

    if final_month not in dic_months:
        print("No valid date was found for " + txt)
        # return "", "", ""
        return None
    else:
        if len(day) == 1:
            day = "0" + day
        real_month = dic_months[final_month]
        # return day, real_month, year
        return date(int(year), int(real_month), int(day))


def find_datetime(txt: [str]) -> Optional[datetime]:
    """
    Utility function to extract date and time from a given string
    :return: a datetime object with the date and time
    """

    day = month = year = hour = minute = second = ""
    index = 0
    final_month = tz = None

    while index < len(date_regex) and final_month not in dic_months:
        match = date_regex[index].search(txt)
        if match:
            match = match[0]

            if date_regex[index] is find_french_date_1:
                _, day, month, year = match.split()

            elif date_regex[index] is find_french_date_2:
                day, month, year = match.split()

            elif date_regex[index] is find_english_date_1:
                _, month, year = match.split(',')
                month, day = month.strip().split()
                month = trad_months[month]

            elif date_regex[index] is find_english_date_2:
                month_day, year = match.split(',')
                month, day = month_day.split(' ')
                month = trad_months[month]

            day, month, year = [x.strip() for x in [day, month, year]]
            final_month = month
            index += 1
        else:
            index += 1

    if final_month not in dic_months:
        print("No valid date was found for " + txt)
        return None
    else:
        if len(day) == 1:
            day = "0" + day
        real_month = dic_months[final_month]
        # return day, real_month, year

    match = time_regex.search(txt)
    file = open("test.txt", "a")
    if match:
        print(f"Found time {match[0]} in {txt}", file=file)
        match = match[0]
        if ":" in match:
            parts = match.split(":")

            if len(parts) == 2:
                hour, minute = parts
                minute = minute.split()[0]
            elif len(parts) == 3:
                hour, minute, second = parts
                second = second.split()[0]
        else:
            hour = match.split()[0]
            minute = "00"
            second = "00"

        if "p" in match.lower():
            hour = str(int(hour) + 12)
        elif "a" in match.lower():
            hour = str(int(hour) + 12)

        hour, minute, second = [x.strip() for x in [hour, minute, second]]

        if "+" in match:
            tz = f"UTC+{match.split('+')[1]}:00"
        elif "-" in match:
            tz = f"UTC-{match.split('-')[1]}:00"

        tz = tzutil.gettz(tz)  # if tz is None, gettz() returns tzlocal()

        if not second:
            second = 0

        try:
            assert int(hour) < 24
            assert int(minute) < 60
            assert int(second) < 60
        except AssertionError:
            raise

        dt = datetime(int(year), int(real_month), int(day), int(hour), int(minute), int(second), tzinfo=tz)
    else:
        print(f"No time found in {txt}", file=file)
        # dt = datetime(int(year), int(real_month), int(day))
        dt = datetime(int(year), int(real_month), int(day), tzinfo=tz)

    file.close()
    return dt


STOP_WORDS = [
    "ans", "faire", "a", "abord", "absolument", "afin", "ah", "ai", "aie", "aient", "aies", "ailleurs",
    "ainsi", "ait",
    "allaient", "allo", "allons", "allô", "alors", "anterieur", "anterieure", "anterieures", "apres", "après",
    "as",
    "assez", "attendu", "au", "aucun", "aucune", "aucuns", "aujourd", "aujourd'hui", "aupres", "auquel",
    "aura", "aurai",
    "auraient", "aurais", "aurait", "auras", "aurez", "auriez", "aurions", "aurons", "auront", "aussi",
    "autant", "autre",
    "autrefois", "autrement", "autres", "autrui", "aux", "auxquelles", "auxquels", "avaient", "avais",
    "avait", "avant",
    "avec", "avez", "aviez", "avions", "avoir", "avons", "ayant", "ayez", "ayons", "b", "bah", "bas", "basee",
    "bat", "beau", "beaucoup",
    "bien", "bigre", "bon", "boum", "bravo", "brrr", "c", "car", "ce", "ceci", "cela", "celle", "celle-ci",
    "celle-là", "celles", "celles-ci",
    "celles-là", "celui", "celui-ci", "celui-là", "celà", "cent", "cependant", "certain", "certaine",
    "certaines", "certains", "certes", "ces",
    "cet", "cette", "ceux", "ceux-ci", "ceux-là", "chacun", "chacune", "chaque", "cher", "chers", "chez",
    "chiche", "chut", "chère", "chères", "ci",
    "cinq", "cinquantaine", "cinquante", "cinquantième", "cinquième", "clac", "clic", "combien", "comme",
    "comment", "comparable", "comparables", "compris", "concernant",
    "contre", "couic", "crac", "d", "da", "dans", "de", "debout", "dedans", "dehors", "deja", "delà",
    "depuis", "dernier", "derniere", "derriere", "derrière", "des",
    "desormais", "desquelles", "desquels", "dessous", "dessus", "deux", "deuxième", "deuxièmement", "devant",
    "devers", "devra", "devrait",
    "different", "differentes", "differents", "différent", "différente", "différentes", "différents", "dire",
    "directe", "directement",
    "dit", "dite", "dits", "divers", "diverse", "diverses", "dix", "dix-huit", "dix-neuf", "dix-sept",
    "dixième", "doit", "doivent",
    "donc", "dont", "dos", "douze", "douzième", "dring", "droite", "du", "duquel", "durant", "dès", "début",
    "désormais", "e", "effet",
    "egale", "egalement", "egales", "eh", "elle", "elle-même", "elles", "elles-mêmes", "en", "encore",
    "enfin", "entre", "envers",
    "environ", "es", "essai", "est", "et", "etant", "etc", "etre", "eu", "eue", "eues", "euh", "eurent",
    "eus", "eusse", "eussent",
    "eusses", "eussiez", "eussions", "eut", "eux", "eux-mêmes", "exactement", "excepté", "extenso",
    "exterieur", "eûmes", "eût", "eûtes",
    "f", "fais", "faisaient", "faisant", "fait", "faites", "façon", "feront", "fi", "flac", "floc", "fois",
    "font", "force", "furent", "fus", "fusse",
    "fussent", "fusses", "fussiez", "fussions", "fut", "fûmes", "fût", "fûtes", "g", "gens", "h", "ha",
    "haut", "hein", "hem", "hep", "hi", "ho", "holà",
    "hop", "hormis", "hors", "hou", "houp", "hue", "hui", "huit", "huitième", "hum", "hurrah", "hé", "hélas",
    "i", "ici", "il", "ils", "importe", "j", "je",
    "jusqu", "jusque", "juste", "k", "l", "la", "laisser", "laquelle", "las", "le", "lequel", "les",
    "lesquelles", "lesquels", "leur", "leurs", "longtemps",
    "lors", "lorsque", "lui", "lui-meme", "lui-même", "là", "lès", "m", "ma", "maint", "maintenant", "mais",
    "malgre", "malgré", "maximale", "me", "meme",
    "memes", "merci", "mes", "mien", "mienne", "miennes", "miens", "mille", "mince", "mine", "minimale",
    "moi", "moi-meme", "moi-même", "moindres", "moins",
    "mon", "mot", "moyennant", "multiple", "multiples", "même", "mêmes", "n", "na", "naturel", "naturelle",
    "naturelles", "ne", "neanmoins", "necessaire",
    "necessairement", "neuf", "neuvième", "ni", "nombreuses", "nombreux", "nommés", "non", "nos", "notamment",
    "notre", "nous", "nous-mêmes", "nouveau",
    "nouveaux", "nul", "néanmoins", "nôtre", "nôtres", "o", "oh", "ohé", "ollé", "olé", "on", "ont", "onze",
    "onzième", "ore", "ou", "ouf", "ouias", "oust",
    "ouste", "outre", "ouvert", "ouverte", "ouverts", "o|", "où", "p", "paf", "pan", "par", "parce",
    "parfois", "parle", "parlent", "parler", "parmi",
    "parole", "parseme", "partant", "particulier", "particulière", "particulièrement", "pas", "passé",
    "pendant", "pense", "permet", "personne",
    "personnes", "peu", "peut", "peuvent", "peux", "pff", "pfft", "pfut", "pif", "pire", "pièce", "plein",
    "plouf", "plupart", "plus", "plusieurs",
    "plutôt", "possessif", "possessifs", "possible", "possibles", "pouah", "pour", "pourquoi", "pourrais",
    "pourrait", "pouvait", "prealable",
    "precisement", "premier", "première", "premièrement", "pres", "probable", "probante", "procedant",
    "proche", "près", "psitt", "pu", "puis", "puisque",
    "pur", "pure", "q", "qu", "quand", "quant", "quant-à-soi", "quanta", "quarante", "quatorze", "quatre",
    "quatre-vingt", "quatrième", "quatrièmement", "que",
    "quel", "quelconque", "quelle", "quelles", "quelqu'un", "quelque", "quelques", "quels", "qui",
    "quiconque", "quinze", "quoi", "quoique", "r", "rare",
    "rarement", "rares", "relative", "relativement", "remarquable", "rend", "rendre", "restant", "reste",
    "restent", "restrictif", "retour", "revoici",
    "revoilà", "rien", "s", "sa", "sacrebleu", "sait", "sans", "sapristi", "sauf", "se", "sein", "seize",
    "selon", "semblable", "semblaient", "semble",
    "semblent", "sent", "sept", "septième", "sera", "serai", "seraient", "serais", "serait", "seras", "serez",
    "seriez", "serions", "serons", "seront",
    "ses", "seul", "seule", "seulement", "si", "sien", "sienne", "siennes", "siens", "sinon", "six",
    "sixième", "soi", "soi-même", "soient", "sois",
    "soit", "soixante", "sommes", "son", "sont", "sous", "souvent", "soyez", "soyons", "specifique",
    "specifiques", "speculatif", "stop", "strictement",
    "subtiles", "suffisant", "suffisante", "suffit", "suis", "suit", "suivant", "suivante", "suivantes",
    "suivants", "suivre", "sujet", "superpose",
    "sur", "surtout", "t", "ta", "tac", "tandis", "tant", "tardive", "te", "tel", "telle", "tellement",
    "telles", "tels", "tenant", "tend", "tenir",
    "tente", "tes", "tic", "tien", "tienne", "tiennes", "tiens", "toc", "toi", "toi-même", "ton", "touchant",
    "toujours", "tous", "tout", "toute",
    "toutefois", "toutes", "treize", "trente", "tres", "trois", "troisième", "troisièmement", "trop", "très",
    "tsoin", "tsouin", "tu", "té", "u", "un",
    "une", "unes", "uniformement", "unique", "uniques", "uns", "v", "va", "vais", "valeur", "vas", "vers",
    "via", "vif", "vifs", "vingt", "vivat",
    "vive", "vives", "vlan", "voici", "voie", "voient", "voilà", "voire", "vont", "vos", "votre", "vous",
    "vous-mêmes", "vu", "vé", "vôtre", "vôtres",
    "w", "x", "y", "z", "zut", "à", "â", "ça", "ès", "étaient", "étais", "était", "étant", "état", "étiez",
    "étions", "été", "étée", "étées", "étés", "êtes", "être", "ô",
    "a", "an", "and", "are", "as", "at", "be", "but", "by", "for", "from", "has", "her", "his", "if", "in",
    "into", "is", "it", "no",
    "not", "of", "on", "or", "such", "that", "the", "their", "then", "there", "these", "they", "this", "to",
    "was", "will", "with"
]




import hashlib
import io
import re
import zipfile
from collections import defaultdict

from typing import List

import yaml

from ..models import Pivot, TransformerOutput
from ..transformers.transformer import Transformer


class MarkdownTransformer(Transformer):
    def __init__(self):
        super().__init__()
        self.output_type = "zip"
        self.output = TransformerOutput(data=None, output=self.output_type,
                                        filename=f'{self.name}_output.{self.output_type}')
        self.seen_names = set()
        self.stats = None
        self.stats_output = ""

    def generate_markdown(self, pivot: Pivot):
        # Générer le contenu du fichier markdown
        frontmatter = {
            "journal": clean_string(pivot.journal_clean),
            "auteur": clean_string(pivot.auteur),
            "titre": clean_string(pivot.titre),
            "date": pivot.date,
            "langue": clean_string(pivot.langue),
            "tags": [clean_string(tag) for tag in pivot.keywords],
            "journal_charts": "journal_" + clean_string(pivot.journal_clean),
            "auteur_charts": "auteur_" + clean_string(pivot.journal_clean),
            "url": pivot.url,
        }

        markdown_content = f"---\n{yaml.dump(frontmatter)}---\n\n{pivot.texte}"

        # Nom du fichier markdown
        # Si le titre est trop long, on le tronque à 100 caractères
        # Si le titre est vide (une fois nettoyé), on utilise le hash du texte
        base_nom = frontmatter["titre"][:100].strip("_") or hashlib.md5(pivot.texte.encode()).hexdigest()

        nom = f"{frontmatter['journal']}/{base_nom}.md"
        if nom in self.seen_names:
            # Si le nom existe déjà, on ajoute la date à la fin (sans l'heure)
            nom = f"{nom[:-3]}_{clean_string(pivot.date).split('t')[0]}.md"
        self.seen_names.add(nom)

        return nom, markdown_content

    def transform(self, pivots: List[Pivot]) -> TransformerOutput:
        self.compute_stats(pivots)
        in_memory_zip = io.BytesIO()
        with zipfile.ZipFile(in_memory_zip, "w", zipfile.ZIP_DEFLATED) as zipf:
            for pivot in pivots:
                filename, content = self.generate_markdown(pivot)
                zipf.writestr(filename, content)
            zipf.writestr("Statistiques.md", self.make_waffle())

        in_memory_zip.seek(0)
        self.output.data = in_memory_zip.getvalue()
        return self.output

    def compute_stats(self, pivots, key="journal"):
        articles_par_cle = defaultdict(int)

        # Parcourez la liste d'éléments pivots
        for pivot in pivots:
            # Incrémentez le compteur pour le journal actuel
            articles_par_cle[pivot.journal_clean] += 1

        # Convertissez le defaultdict en un dictionnaire Python standard pour la sortie
        self.stats = dict(articles_par_cle)

    def make_waffle(self):
        output = "## Articles par journal \n\n" \
                 "```chartsview\n"
        chart = {
            "type": "Treemap",
            "data": {
                "name": "root",
                "children": [],
            },
            "options": {
                "colorField": "name",
                "enableSearchInteraction": {
                    "field": "journal_chart"
                }
            }
        }
        for journal, value in self.stats.items():
            # do this to avoid searching in the Statistics.md file
            search_value = 'journal_' + clean_string(journal) + '" -file:(Statistiques) "'
            chart['data']['children'].append({'name': journal, 'value': value, 'journal_chart': search_value})

        output += yaml.dump(chart)
        output += "```"
        return output

def clean_string(s):
    # Fonction pour nettoyer les chaînes de caractères
    s = re.sub(r"[^\w\s]", "", s)  # Supprimer les caractères spéciaux
    s = s.lower()  # Mettre en minuscule
    s = s.strip()  # Supprimer les espaces au début et à la fin
    s = re.sub(r'\s+', '_', s)  # Remplacer les espaces par des underscores
    return s



import json
import logging
import os
from pathlib import Path
from abc import ABC, abstractmethod
from typing import List, Optional, Any
import re

import unicodedata

from ..models import Error, Pivot, TransformerOutput, Params

# Transformer initialization, allows all transformers to access the output path end prevents to set it multiple times

output_path = os.getenv("EUROPARSER_OUTPUT", None)

if output_path is None:
    logging.warning("EUROPARSER_OUTPUT not set, disabling output")
else:
    output_path = Path(output_path)

    if not output_path.is_dir():
        logging.warning(f"Output path {output_path} is not a directory, disabling output")
        output_path = None

if output_path:
    output_path.mkdir(parents=True, exist_ok=True)


class Transformer(ABC):
    output_path = output_path

    def __init__(self, params: Optional[Params] = None, **kwargs: Optional[Any]):
        self.name: str = type(self).__name__.split('Transformer')[0].lower()
        self.errors: List[Error] = []
        self._logger = logging.getLogger(self.name)
        self._logger.setLevel(logging.WARNING)
        # self.output_type = "json" # TODO any use of setting the output type ? Should maybe be a None ?
        self.params = params or Params(**kwargs)  # If no kwargs are passed, params will be initialized with default values

    @abstractmethod
    def transform(self, pivot: List[Pivot]) -> TransformerOutput:
        """
        Returns the transformed data, the output_type, and the output_filename
        """
        raise NotImplementedError()

    def _add_error(self, error, article):
        self.errors.append(Error(message=str(error), article=article.text, transformer=self.name))

    def _persist_errors(self, filename):
        """
        Save all errors to disk
        :param filename: name of the file being transformed
        """
        dir_path = Path(os.path.join(str(Path.home()), 'europarser'))
        dir_path.mkdir(parents=True, exist_ok=True)
        path = os.path.join(dir_path, f"errors-{filename}.json")
        mode = "a" if os.path.exists(path) else "w"
        with open(path, mode, encoding="utf-8") as f:
            json.dump([e.dict() for e in self.errors], f, ensure_ascii=False)

    @staticmethod
    def _format_value(value: str):
        # value = re.sub(r"[éèê]", "e", value)
        # value = re.sub(r"ô", "o", value)
        # value = re.sub(r"à", "a", value)
        # value = re.sub(r"œ", "oe", value)
        # value = re.sub(r"[ïîì]", "i", value)
        value = strip_accents(value)
        value = re.sub(r"""[-\[\]'":().=?!,;<>«»—^*\\/|]""", ' ', value)
        return ''.join([w.capitalize() for w in value.split(' ')])


def strip_accents(s):
    return ''.join(c for c in unicodedata.normalize('NFKD', s) if unicodedata.category(c) != 'Mn')



import json
import re
import time
import zipfile
from datetime import datetime, date
from typing import List, Optional, Any
from io import StringIO, BytesIO

import plotly.express as px
import plotly.graph_objects as go
import plotly.io as pio
import polars as pl

from ..models import Pivot, TransformerOutput, Params
from ..transformers.transformer import Transformer

# locale.setlocale(locale.LC_ALL, "fr_FR.UTF-8")
pio.templates.default = "none"


class StatsTransformer(Transformer):
    mois = ("janvier", "février", "mars", "avril", "mai", "juin", "juillet", "août", "septembre", "octobre", "novembre",
            "décembre")
    COOL_COLORS = {"bluered", "plasma", "plotly3", "rainbow", "portland", "spectral", "inferno", "sunsetdark", "matter",
                   "turbo", "rdbu"}
    COOL_COLORS |= {e + "_r" for e in COOL_COLORS}
    MAIN_COLOR = "bluered_r"

    @staticmethod
    def clean(s: str) -> str:
        return re.sub(r"\s+", " ", s).strip()

    @staticmethod
    def to_date(s: str) -> date:
        return datetime.strptime(s, "%Y %m %d").date()

    @staticmethod
    def to_monthyear(d: datetime) -> str:
        return f"{d.year}-{d.month:02}"

    @staticmethod
    def int_to_datetime(i: int) -> datetime:
        return datetime.fromtimestamp(i)

    @staticmethod
    def int_to_monthyear(i: int) -> str:
        # return self.to_monthyear(self.int_to_datetime(i))
        dt = datetime.fromtimestamp(i)
        return f"{dt.year}-{dt.month:02}"

    @staticmethod
    def int_to_monthyear_intversion( i: int) -> int:
        dt = datetime.fromtimestamp(i)
        return dt.year * 100 + dt.month  # --> int(f"{dt.year}{dt.month:02}") equivalent

    @staticmethod
    def for_display(mois_int: int) -> str:
        return f"{mois_int // 100}-{mois_int % 100:02}"
        # ## TODO : compare performance with this
        # mois_str = str(mois_int)
        # return f"{mois_str[:-2]}-{mois_str[-2:]}"

    def __init__(self, params: Optional[Params] = None, **kwargs: Optional[Any]):
        super().__init__(params, **kwargs)
        self.df = None
        self.data = None
        self.res = None

        self.stats_done = False
        self.processed_stats_done = False
        self.plots_done = False

        self.pivot_list = None

        self.output_type = {
            "stats": "json",
            "processed_stats": "json",
            "plots": "zip",
        }

        self.output = {
            k: TransformerOutput(
                data=None,
                output=self.output_type[k],
                filename=f'{k}_output.{self.output_type[k]}'
            )
            for k in self.output_type
        }

    def transform(self, pivot_list: List[Pivot]) -> TransformerOutput:
        self._logger.debug("Starting to compute stats")
        t1 = time.time()

        self.pivot_list = pivot_list
        self.df = pl.from_records([p.model_dump() for p in self.pivot_list])

        self.df = self.df.with_columns(
            pl.col('journal_clean').str.strip_chars().alias('journal_clean'),

            pl.col('epoch').map_elements(self.int_to_monthyear_intversion).alias('mois'),

            pl.col('keywords')
            .str.replace_all(r"[(\[\])']", "")
            .str.split(',')
            .list.eval(pl.element().filter(pl.element() != ""))
            .list.eval(pl.element().str.strip_chars(" ,\n\t"))
            .list.drop_nulls(),
        ).with_row_count()

        self.list_mois = list(self.df.select("mois").to_series().unique().map_elements(self.for_display))

        self.data = {
            "journal": (
                self.df
                .group_by("journal_clean")
                .agg(pl.col("row_nr").agg_groups())
                .sort("journal_clean")
                .select(pl.col("journal_clean").alias("journal"), pl.col("row_nr").alias("index_list"))
            ),
            "mois": (
                self.df
                .group_by("mois")
                .agg(pl.col("row_nr").agg_groups())
                .sort("mois")
                .select(pl.col("mois").alias("mois"), pl.col("row_nr").alias("index_list"))
                .with_columns(pl.col("mois").map_elements(self.for_display))
            )
            ,
            "auteur": (
                self.df
                .group_by("auteur")
                .agg(pl.col("row_nr").agg_groups())
                .sort("auteur")
                .select(pl.col("auteur").alias("auteur"), pl.col("row_nr").alias("index_list"))
            ),
            "mot_cle": (
                self.df
                .explode("keywords")
                .drop_nulls()
                .group_by("keywords")
                .agg(pl.col("row_nr").agg_groups())
                .sort("keywords")
                .select(pl.col("keywords").alias("mot_cle"), pl.col("row_nr").alias("index_list"))
            ),

            "mois_journal": (
                self.df
                .group_by(["mois", "journal_clean"])
                .agg(pl.col("row_nr").agg_groups())
                .sort(["journal_clean", "mois"])
                .select(pl.col("journal_clean").alias("journal"), pl.col("mois").alias("mois"),
                        pl.col("row_nr").alias("index_list"))
                .with_columns(
                    pl.col("mois").map_elements(self.for_display)
                )
            ),
            "mois_kw": (
                self.df
                .explode("keywords")
                .group_by(["mois", "keywords"])
                .agg(pl.col("row_nr").agg_groups())
                .sort(["mois", "keywords"])
                .select(pl.col("mois").alias("mois"), pl.col("keywords").alias("mot_cle"),
                        pl.col("row_nr").alias("index_list"))
                .with_columns(
                    pl.col("mois").map_elements(self.for_display)
                )
            ),
            "mois_auteur": (
                self.df
                .group_by(["mois", "auteur"])
                .agg(pl.col("row_nr").agg_groups())
                .sort(["auteur", "mois"])
                .select(pl.col("auteur").alias("auteur"), pl.col("mois").alias("mois"),
                        pl.col("row_nr").alias("index_list"))
                .with_columns(
                    pl.col("mois").map_elements(self.for_display)
                )
            ),
        }

        self.res = {
            key: {
                key2: val2
                for key2, val2 in list(zip(*val.to_dict(as_series=False).values()))
            }
            for key, val in self.data.items() if len(val.columns) == 2
        }

        self.res.update({
            key: {
                f"{key2}_{key2_bis}": val2
                for key2, key2_bis, val2 in list(zip(*val.to_dict(as_series=False).values()))
            }
            for key, val in self.data.items() if len(val.columns) == 3
        })

        self._logger.debug(f"Time to compute stats: {time.time() - t1:.2f}s")
        self.stats_done = True

        self._transform_processed()

        self.output["stats"].data = json.dumps(self.res)
        return self.output["stats"]

    def _transform_processed(self, *args, **kwargs):
        if not self.stats_done:
            raise ValueError("You must compute the stats before generating the processed ones")

        self.processed_stats = {
            "journal":
                (
                    self.data["journal"]
                    .select("journal", pl.col("index_list").map_elements(lambda x: len(x)).alias("count"))
                    .filter(pl.col("count") >= self.params.minimal_support_journals)
                    .sort("count", descending=True)
                ),
            "mois":
                (
                    self.data["mois"]
                    .select("mois", pl.col("index_list").map_elements(lambda x: len(x)).alias("count"))
                    .filter(pl.col("count") >= self.params.minimal_support_dates)
                    .sort("mois")
                ),
            "auteur":
                (
                    self.data["auteur"]
                    .select(pl.col("auteur").cast(pl.Utf8), pl.col("index_list").map_elements(lambda x: len(x)).alias("count"))
                    .sort("count", descending=True)
                    .filter(pl.col("count") >= self.params.minimal_support_authors)
                    .filter(pl.col("auteur") != "Unknown")
                ),
            "mot_cle":
                (
                    self.data["mot_cle"]
                    .select("mot_cle", pl.col("index_list").map_elements(lambda x: len(x)).alias("count"))
                    .filter(pl.col("count") >= self.params.minimal_support_kw)
                    .sort("count", descending=True)
                ),
        }

        self.journal_order = self.processed_stats["journal"].select(pl.col("journal")).to_series().to_list()
        self.auteur_order = self.processed_stats["auteur"].select(pl.col("auteur")).to_series().to_list()
        self.mot_cle_order = self.processed_stats["mot_cle"].select(pl.col("mot_cle")).to_series().to_list()

        self.processed_stats.update({
            "mois_journal": {
                journal: (
                    self.data["mois_journal"]
                    .filter(pl.col("journal") == journal)
                    .select("mois", pl.col("index_list").map_elements(lambda x: len(x)).alias("count"))
                    .sort("mois")
                ) for journal in self.journal_order
            },
            "mois_kw": {
                kw: (
                    self.data["mois_kw"]
                    .filter(pl.col("mot_cle") == kw)
                    .select("mois", pl.col("index_list").map_elements(lambda x: len(x)).alias("count"))
                    .sort("mois")
                ) for kw in self.mot_cle_order
            },
            "mois_auteur": {
                auteur: (
                    self.data["mois_auteur"]
                    .filter(pl.col("auteur") == auteur)
                    .select("mois", pl.col("index_list").map_elements(lambda x: len(x)).alias("count"))
                    .sort("mois")
                ) for auteur in self.auteur_order
            },
        })

        self.processed_stats_done = True

    def get_stats(self, *args, **kwargs):
        if not self.stats_done:
            raise ValueError("You must compute the stats before getting them")
        return self.output["stats"]

    def get_processed_stats(self, *args, **kwargs):
        if not self.stats_done or not self.processed_stats_done:
            raise ValueError("You must compute the stats before getting the processed ones")

        self.output["processed_stats"].data = json.dumps({
            "journal": {k: v for k, v in zip(*self.processed_stats["journal"].to_dict(as_series=False).values())},
            "mois": {k: v for k, v in zip(*self.processed_stats["mois"].to_dict(as_series=False).values())},
            "auteur": {k: v for k, v in zip(*self.processed_stats["auteur"].to_dict(as_series=False).values())},
            "mois_journal": {
                journal: {k: v for k, v in zip(*df.to_dict(as_series=False).values())} for journal, df in self.processed_stats["mois_journal"].items()
            },
            "mois_kw": {
                kw: {k: v for k, v in zip(*df.to_dict(as_series=False).values())} for kw, df in self.processed_stats["mois_kw"].items()
            },
            "mois_auteur": {
                auteur: {k: v for k, v in zip(*df.to_dict(as_series=False).values())} for auteur, df in self.processed_stats["mois_auteur"].items()
            },
        })

        return self.output["processed_stats"]

    def get_plots(self, *args, **kwargs):
        if not self.stats_done or not self.processed_stats_done:
            raise ValueError("You must compute the stats before generating the plots")

        if self.plots_done:
            return self.output["plots"]

        self._logger.debug("Starting to compute plots")
        t1 = time.time()

        with BytesIO() as zip_io:
            with zipfile.ZipFile(zip_io, mode="w", compression=zipfile.ZIP_DEFLATED) as zip_file:
                self._get_plots(zip_file)

            self._logger.debug(f"Time to compute plots: {time.time() - t1:.2f}s")
            self.output["plots"].data = zip_io.getvalue()

            self.plots_done = True
            return self.output["plots"]

    def _get_plots(self, zip_file):
        self.zip_file = zip_file

        self._get_plots_journal()
        self._get_plots_mois()
        self._get_plots_auteur()
        self._get_plots_mot_cle()

        self._get_plots_mois_journal()
        self._get_plots_mois_kw()
        self._get_plots_mois_auteur()

    def _get_plots_journal(self):
        fig = px.bar(
            self.processed_stats["journal"],
            x="journal",
            y="count",
            color="count",
            labels={"x": "Journal", "y": "Nombre d'articles", "count": "Nombre d'articles"},
            title="Nombre d'articles par journal",
            color_continuous_scale=self.MAIN_COLOR,
        )

        fig.update_layout(
            xaxis_tickformat="%B %Y",
        )

        self.zip_file.writestr("journal.html", fig.to_html())

    def _get_plots_mois(self):
        fig = px.bar(
            self.processed_stats["mois"],
            x="mois",
            y="count",
            color="count",
            labels={"x": "Mois", "y": "Nombre d'articles", "count": "Nombre d'articles"},
            title="Nombre d'articles par mois",
            color_continuous_scale=self.MAIN_COLOR,
        )
        fig.update_layout(
            xaxis_tickformat="%B %Y",
        )
        self.zip_file.writestr("mois.html", fig.to_html())

    def _get_plots_auteur(self):
        fig = px.bar(
            self.processed_stats["auteur"],
            x="auteur",
            y="count",
            color="count",
            labels={"x": "Auteur", "y": "Nombre d'articles", "count": "Nombre d'articles"},
            title="Nombre d'articles par auteur",
            color_continuous_scale=self.MAIN_COLOR,
        )
        self.zip_file.writestr("auteur.html", fig.to_html())

    def _get_plots_mot_cle(self):
        fig = px.bar(
            self.processed_stats["mot_cle"],
            x="mot_cle",
            y="count",
            color="count",
            labels={"x": "Mot clé", "y": "Nombre d'articles", "count": "Nombre d'articles"},
            title="Nombre d'articles par mot clé",
            color_continuous_scale=self.MAIN_COLOR,
        )
        self.zip_file.writestr("mot_cle.html", fig.to_html())

    def _get_plots_mois_journal(self):
        fig = px.line()

        for journal in self.journal_order:
            df = self.processed_stats["mois_journal"][journal]

            if len(df) < self.params.minimal_support:
                continue

            fig.add_trace(
                go.Scatter(
                    x=df.select("mois").to_series(),
                    y=df.select("count").to_series(),
                    name=journal,
                    connectgaps=True,
                )
            )

        fig.update_layout(
            xaxis_tickformat="%B %Y",
            title="Nombre d'articles par mois et par journal",
            xaxis_title="Mois",
            yaxis_title="Nombre d'articles",
        )

        self.zip_file.writestr("mois_journal.html", fig.to_html())

    def _get_plots_mois_kw(self):
        fig = px.line()

        for kw in self.mot_cle_order:
            df = self.processed_stats["mois_kw"][kw]

            if len(df) < self.params.minimal_support:
                continue

            fig.add_trace(
                go.Scatter(
                    x=df.select("mois").to_series(),
                    y=df.select("count").to_series(),
                    name=kw,
                    connectgaps=True,
                )
            )

        fig.update_layout(
            xaxis_tickformat="%B %Y",
            title="Nombre d'articles par mois et par mot clé",
            xaxis_title="Mois",
            yaxis_title="Nombre d'articles",
            margin=dict(l=20),
        )

        self.zip_file.writestr("mois_kw.html", fig.to_html())

    def _get_plots_mois_auteur(self):
        fig = px.line()

        for auteur in self.auteur_order:
            df = self.processed_stats["mois_auteur"][auteur]

            if len(df) < self.params.minimal_support:
                continue

            fig.add_trace(
                go.Scatter(
                    x=df.select("mois").to_series(),
                    y=df.select("count").to_series(),
                    name=auteur,
                    connectgaps=True,
                )
            )

        fig.update_layout(
            xaxis_tickformat="%B %Y",
            title="Nombre d'articles par mois et par auteur",
            xaxis_title="Mois",
            yaxis_title="Nombre d'articles",
        )

        self.zip_file.writestr("mois_auteur.html", fig.to_html())


if __name__ == '__main__':
    import cProfile
    import pstats

    for mode in ["small", "medium", "large"]:
        with open(f"../../profiler/data/pivots{f'_{mode}' if mode else ''}.json", "r", encoding="utf-8") as f:
            dict_ = json.load(f)
            dict_ = list(dict_.values())
            pivot_list = [Pivot(**d) for d in dict_]

        transformer = StatsTransformer()

        pr = cProfile.Profile()
        pr.enable()

        res = transformer.transform(pivot_list)
        zip_file = transformer.get_plots()

        pr.disable()
        s = StringIO()
        sortby = 'cumulative'
        ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
        # ps.print_stats()

        with open(f"../../profiler/results/{mode}/stats.json", "w") as f:
            json.dump(res, f, indent=4)

        with open(f"../../profiler/results/{mode}/profiler.tsv", "w") as f:
            profiler_res = s.getvalue().splitlines()

            # print(f"{mode = }\n\tprofiler => {profiler_res[0].strip()}")

            f.write("\n".join(profiler_res[4:]))

        with open(f"../../profiler/results/{mode}/plots.zip", "wb") as f:
            f.write(zip_file)



from argparse import ArgumentParser
from pathlib import Path

parser = ArgumentParser()

parser.add_argument("-c", "--cli", help="Run as CLI", action="store_true")
parser.add_argument("-f", "--folder", help="Folder to process", type=str)
parser.add_argument("-o", "--output", help="Outputs to generate", nargs="+", type=str)
parser.add_argument("--support-kw", help="Minimal support for keywords", type=int)
parser.add_argument("--support-authors", help="Minimal support for authors", type=int)
parser.add_argument("--support-journals", help="Minimal support for journals", type=int)
parser.add_argument("--support-dates", help="Minimal support for dates", type=int)
parser.add_argument("--support", help="Minimal support for all", type=int)
parser.add_argument("--filter-keywords", help="Filter keywords", type=bool, default=False)
parser.add_argument("--filter-lang", help="Filter language", type=bool, default=False)

parser.add_argument("--api", help="Run as API", action="store_true")
parser.add_argument("--host", help="Host to bind to", type=str)
parser.add_argument("--port", help="Port to bind to", type=int)
parser.add_argument(
    "--expose",
    help="Expose the API (shorthand for --host 0.0.0.0 --port 8000), dosent override --host or --port if specified",
    action="store_true",
)


if __name__ == '__main__':
    args = parser.parse_args()

    if args.api or all([args.host, args.port]) or args.expose:
        import uvicorn
        from europarser.api import app

        if args.expose:
            args.host = args.host or "0.0.0.0"

        uvicorn.run(app, host=args.host or "127.0.0.1", port=args.port or 8000)
        exit(0)

    if args.cli:
        if not all([args.folder, args.output]):
            print("You need to specify a input folder and outputs")
            parser.print_help()
            exit(1)

    if not all([args.folder, args.output]):
        parser.print_help()
        exit(1)

    from europarser import main
    from europarser import Params, Output
    from typing import get_args

    possible_outputs = get_args(Output)

    folder = Path(args.folder)
    assert folder.is_dir(), f"Folder {folder} does not exist"
    outputs = args.output
    for output in outputs:
        assert output in possible_outputs, f"Output {output} is not supported"

    params = Params(
        minimal_support_kw=args.support_kw,
        minimal_support_authors=args.support_authors,
        minimal_support_journals=args.support_journals,
        minimal_support_dates=args.support_dates,
        minimal_support=args.support or 1,
        filter_keywords=args.filter_keywords,
        filter_lang=args.filter_lang
    )

    main(folder, outputs, params=params)
    exit(0)



import io
import logging
import os
import zipfile
from enum import Enum
from pathlib import Path
from typing import List, Annotated  # , Optional

from fastapi import FastAPI, UploadFile, Request, Form, HTTPException, File, Depends  # , Query
from fastapi.staticfiles import StaticFiles
from fastapi.responses import StreamingResponse, HTMLResponse
from fastapi.templating import Jinja2Templates

from ..models import FileToTransform, TransformerOutput, Params  # , Output
from .. import pipeline
from .utils import get_mimetype

root_dir = os.path.dirname(__file__)
host = os.getenv('EUROPARSER_SERVER', '')
app = FastAPI()

app.mount("/static", StaticFiles(directory=os.path.join(root_dir, "static")), name="static")
templates = Jinja2Templates(directory=os.path.join(root_dir, "templates"))

logger = logging.getLogger("europarser_api.api")
logger.setLevel(logging.DEBUG)


class Outputs(str, Enum):
    json = "json"
    txm = "txm"
    iramuteq = "iramuteq"
    gephi = "gephi"
    csv = "csv"
    excel = "excel"
    stats = "stats"
    processed_stats = "processed_stats"
    plots = "plots"
    markdown = "markdown"


@app.get("/", response_class=HTMLResponse)
async def read_root(request: Request):
    return templates.TemplateResponse('new_wui.html', {'request': request, 'host': host})


@app.post("/upload")
async def handle_files(files: Annotated[list[UploadFile], File(...)],
                       output: Annotated[list[Outputs], Form(...)],
                       params: Annotated[Params, Depends()]):
    if len(files) == 1 and files[0].filename == "":
        raise HTTPException(status_code=400, detail="No File Provided")
    # parse all files
    try:
        to_process = [FileToTransform(name=f.filename, file=f.file.read().decode('utf-8')) for f in files]
    except UnicodeDecodeError:
        raise HTTPException(status_code=400, detail="Invalid File Provided")

    # process result
    results: list[TransformerOutput] = pipeline(to_process, output, params)

    # if only one output was required let's return a single file
    if len(results) == 1:
        result = results[0]

        if isinstance(result.data, io.StringIO) or isinstance(result.data, io.BytesIO):
            pass
        elif not isinstance(result.data, bytes):
            result.data = io.StringIO(result.data)
        else:
            result.data = io.BytesIO(result.data)

        return StreamingResponse(
            result.data,
            media_type=get_mimetype(result.output),
            headers={'Content-Disposition': f"attachment; filename={result.filename}"}
        )

    # else let's create a zip with all files
    zip_io = io.BytesIO()
    with zipfile.ZipFile(zip_io, mode='w', compression=zipfile.ZIP_DEFLATED) as temp_zip:
        for result in results:
            logger.info(f"Adding {result.filename} to zip")
            if result.output == "zip":
                name = Path(result.filename).stem  # get filename without extension (remove .zip basically)
                logger.info(f"Zip file detected, extracting {name}")
                with zipfile.ZipFile(io.BytesIO(result.data), mode='r') as z:
                    for f in z.namelist():
                        temp_zip.writestr(f"{name}/{f}", z.read(f))
                continue

            temp_zip.writestr(f"{result.filename}", result.data)

    zip_io.seek(0)
    return StreamingResponse(
        zip_io,
        media_type="application/zip",
        headers={'Content-Disposition': 'attachment; filename=result.zip'}
    )

def main():
    from argparse import ArgumentParser

    import uvicorn

    parser = ArgumentParser()
    parser.add_argument("--host", default="127.0.0.1", help="Host to bind to")
    parser.add_argument("--port", default=8000, help="Port to bind to", type=int)
    args = parser.parse_args()
    uvicorn.run(app, host=args.host, port=args.port)

if __name__ == "__main__":
    main()



import io
import re
import xml.dom.minidom as dom
from xml.sax.saxutils import quoteattr, escape
from typing import List

from ..models import Pivot, TransformerOutput
from ..transformers.transformer import Transformer


class TXMTransformer(Transformer):
    def __init__(self):
        super(TXMTransformer, self).__init__()
        self.output_type = "xml"
        self.output = TransformerOutput(data=None, output=self.output_type,
                                        filename=f'{self.name}_output.{self.output_type}')

    def transform(self, pivot_list: List[Pivot]) -> TransformerOutput:
        with io.StringIO() as f:
            f.write("<corpus>")

            for pivot in pivot_list:
                pivot_dict = {
                    "titre": pivot.titre,
                    "date": pivot.date,
                    "journal": pivot.journal,
                    "auteur": pivot.auteur,
                    "annee": pivot.annee,
                    "mois": pivot.mois,
                    "jour": pivot.jour,
                    "journal_clean": pivot.journal_clean,
                    "keywords": ', '.join(pivot.keywords),
                    "langue": pivot.langue,
                    "url": pivot.url,
                }
                f.write("<text ")
                for key, value in pivot_dict.items():
                    if value:
                        f.write(f"{key}={quoteattr(str(value))} ")
                f.write(f"> {escape(pivot.texte.strip()) if pivot.texte else ''} </text>\n")

            f.write("</corpus>")
            self.output.data = dom.parseString(f.getvalue()).toprettyxml()
            return self.output



from io import BytesIO
from typing import List

from ..models import Pivot, TransformerOutput
from ..transformers.transformer import Transformer

import pandas as pd
from pandas import ExcelWriter


class ExcelTransformer(Transformer):
    def __init__(self):
        super(ExcelTransformer, self).__init__()
        self.output_type = "excel"
        self.output = TransformerOutput(data=None, output=self.output_type,
                                        filename=f'{self.name}_output.xlsx')

    def transform(self, pivot_list: List[Pivot]) -> TransformerOutput:
        df = pd.DataFrame.from_records([p.model_dump() for p in pivot_list])
        with BytesIO() as output:
            df.to_excel(output, index=False)
            output.seek(0)
            self.output.data = output.read()
        return self.output




# SPDX-FileCopyrightText: 2024-present Marceau <git@marceau-h.fr>
#
# SPDX-License-Identifier: AGPL-3.0-or-later
from .models import FileToTransform, Params, Output, OutputFormat
from .pipeline import pipeline
from .main import main

__all__ = ["FileToTransform", "Params", "Output", "OutputFormat", "pipeline", "main"]



import io
from typing import List

from ..models import Pivot, TransformerOutput
from ..transformers.transformer import Transformer


class IramuteqTransformer(Transformer):
    banned_keys = {"texte", "complement", "date", "epoch"}

    def __init__(self):
        super(IramuteqTransformer, self).__init__()
        self.output_type = "txt"
        self.output = TransformerOutput(data=None, output=self.output_type,
                                        filename=f'{self.name}_output.{self.output_type}')

    def transform(self, pivot_list: List[Pivot]) -> TransformerOutput:
        with io.StringIO() as f:
            for pivot in pivot_list:
                dic = pivot.model_dump(exclude=self.banned_keys)
                f.write(f"""**** {' '.join([f"*{k}_{self._format_value(str(v))}" for k, v in dic.items()])}\n""")
                f.write(pivot.texte)
                f.write('\n\n')
            self.output.data = f.getvalue()
            return self.output



import hashlib
import json
import logging
import concurrent.futures
import re
from collections import Counter
from typing import Optional, Any

from bs4 import BeautifulSoup
from tqdm.auto import tqdm

from .models import FileToTransform, Pivot, Params
from .transformers.transformer import Transformer
from .utils import find_datetime
from .daniel_light import get_KW
from .lang_detect import detect_lang


class BadArticle(Exception):
    pass


class PivotTransformer(Transformer):
    journal_split = re.compile(r"\(| -|,? no. | \d|  | ;|\.fr")
    double_spaces_and_beyond = re.compile(r"(\s{2,})")
    def __init__(self, params: Optional[Params] = None, **kwargs: Optional[Any]):
        super().__init__(params, **kwargs)
        self.corpus = []
        self.bad_articles = []
        self.ids = set()
        self.all_keywords = Counter()

    def subspaces(self, s):
        return self.double_spaces_and_beyond.sub(r"\1", s).strip()

    def transform(self, files_to_transform: list[FileToTransform]) -> list[Pivot]:
        for file in files_to_transform:
            self._logger.debug("Processing file " + file.name)
            soup = BeautifulSoup(file.file, 'lxml')
            articles = soup.find_all("article")

            with concurrent.futures.ThreadPoolExecutor(1) as executor:
                futures = [executor.submit(self.transform_article, article) for article in articles]
                concurrent.futures.wait(futures, return_when=concurrent.futures.ALL_COMPLETED)

        self._logger.info(f"Nombre d'articles : {len(self.corpus)}")

        self.persist_json()
        self.apply_parameters()

        return sorted(self.corpus, key=lambda x: x.epoch)

    def transform_article(self, article):
        try:
            doc = {
                "journal": None,
                "date": None,
                "annee": None,
                "mois": None,
                "jour": None,
                "heure": None,
                "minute": None,
                "seconde": None,
                "epoch": None,
                "titre": None,
                "complement": None,
                "texte": None,
                "auteur": "Unknown",
                "journal_clean": None,
                "keywords": None,
                "langue": "UNK",
                "url": None,
            }
            try:
                doc["journal"] = self.subspaces(article.find("span", attrs={"class": "DocPublicationName"}).text)
            except Exception as e:
                self._logger.debug("pas un article de presse")
                raise BadArticle("journal")

            try:
                doc_header = article.find("span", attrs={"class": "DocHeader"}).text
            except AttributeError:
                doc_header = ""

            try:
                doc_sub_section = article.find("span", attrs={"class": "DocTitreSousSection"}).find_next_sibling("span")
            except AttributeError:
                doc_sub_section = ""

            try:
                datetime = find_datetime(doc_header or doc_sub_section)
            except ValueError:
                raise BadArticle("datetime")

            if datetime:
                doc.update({
                    "date": datetime.strftime("%Y %m %dT%H:%M:%S"),
                    "annee": datetime.year,
                    "mois": datetime.month,
                    "jour": datetime.day,
                    "heure": datetime.hour,
                    "minute": datetime.minute,
                    "seconde": datetime.second,
                    "epoch": datetime.timestamp()
                })

            try:
                doc_titre_full = article.find("div", attrs={"class": "titreArticle"})
                assert doc_titre_full is not None
            except AssertionError:
                try:
                    doc_titre_full = article.find("p", attrs={"class": "titreArticleVisu"})
                    assert doc_titre_full is not None
                except AssertionError:
                    raise BadArticle("titre")

            try:
                doc["titre"] = doc_titre_full.find("p", attrs={
                    "class": "sm-margin-TopNews titreArticleVisu rdp__articletitle"}).text
            except AttributeError:
                try:
                    doc["titre"] = doc_titre_full.find("div", attrs={"class": "titreArticleVisu"}).text
                except AttributeError:
                    try:
                        doc["titre"] = doc_titre_full.text
                    except AttributeError:
                        raise BadArticle("titre")

            doc["titre"] = self.subspaces(doc["titre"])

            try:
                doc_bottomNews = doc_titre_full.find("p", attrs={"class": "sm-margin-bottomNews"}).text
                if not doc_bottomNews:
                    raise AttributeError
            except AttributeError:
                doc_bottomNews = ""

            try:
                doc_subtitle = doc_titre_full.find("p", attrs={"class": "sm-margin-TopNews rdp__subtitle"}).text
                if not doc_subtitle:
                    raise AttributeError
            except AttributeError:
                doc_subtitle = ""

            doc["complement"] = self.subspaces(
                " | ".join((doc_header, doc_sub_section, doc_bottomNews, doc_subtitle))
            )

            try:
                doc_text = article.find("div", attrs={"class": "docOcurrContainer"})
                assert doc_text is not None and doc_text.text.strip()
            except AssertionError:
                if article.find("div", attrs={"class": "DocText clearfix"}) is None:
                    raise BadArticle("texte")
                else:
                    doc_text = article.find("div", attrs={"class": "DocText clearfix"})

            doc["url"] = ""
            for u in doc_text.select("a"):
                if "Cet article est paru dans" in u.get_text():
                    doc["url"] = u.get("href")
                    break


            doc["texte"] = self.subspaces(doc_text.text.strip())

            doc_auteur = doc_titre_full.find_next_sibling('p')

            if doc_auteur and "class" in doc_auteur.attrs and doc_auteur.attrs['class'] == ['sm-margin-bottomNews']:
                doc["auteur"] = self.subspaces(doc_auteur.text.strip().lower())

            # on garde uniquement le titre (sans les fioritures)
            journal_clean = self.journal_split.split(doc["journal"])[0]
            doc["journal_clean"] = self.subspaces(journal_clean)

            doc["keywords"] = get_KW(doc["titre"], doc["texte"])

            self.all_keywords.update(doc["keywords"])

            id_ = ' '.join([doc["titre"], doc["journal_clean"], doc["date"]])

            langue = detect_lang(doc["texte"])
            if langue:
                doc["langue"] = langue

            if id_ not in self.ids:
                self.corpus.append(Pivot(**doc))
                self.ids.add(id_)

        except BadArticle as e:
            if self._logger.isEnabledFor(logging.DEBUG):
                self._add_error(e, article)
                self.bad_articles.append(article)

        return

    def apply_parameters(self):
        if self.params.filter_keywords is True:
            for article in self.corpus:
                article.keywords = list(filter(self.filter_kw, article.keywords))

        return self.corpus

    def filter_kw(self, keyword):
        return self.all_keywords[keyword] > 1

    def get_bad_articles(self):
        print(self.bad_articles)

    def persist_json(self):
        """
        utility function to persist the result of the pivot transformation
        """
        if not self.output_path:
            return

        json_ver = json.dumps({i: article.dict() for i, article in enumerate(self.corpus)}, ensure_ascii=False)
        # hash_json = hashlib.sha256(json_ver.encode()).hexdigest()
        # with (self.output_path / f"{hash_json}.json").open("w", encoding="utf-8") as f:
        output_file = self.output_path / f"{hashlib.sha256(json_ver.encode()).hexdigest()}.json"

        with output_file.open("w", encoding="utf-8") as f:
            f.write(json_ver)


if __name__ == "__main__":
    import cProfile
    import pstats

    from pathlib import Path

    pr = cProfile.Profile()
    pr.enable()

    p = PivotTransformer()

    for file in tqdm(list(Path("/home/marceau/Nextcloud/eurocollectes").glob("**/*.HTML"))):
        with file.open(mode="r", encoding="utf-8") as f:
            p.transform(FileToTransform(file=f.read(), name=file.name))

    p.get_bad_articles()

    pr.disable()
    ps = pstats.Stats(pr).sort_stats('cumulative')
    ps.print_stats()



from pathlib import Path
import requests

from concurrent.futures import ThreadPoolExecutor

from europarser.models import FileToTransform, Params

# server = "https://ceres.huma-num.fr/europarser"
server = "http://localhost:8000"

file = Path("../examples/resources/1.HTML")

output = ["json"]

params = Params()

stress = 32


def send_file(file, output: list[str], params: Params) -> int:
    url = f"{server}/upload"
    data = {'output': output, 'params': params}
    r = requests.post(url, files=file, data=data)
    return r.status_code


if __name__ == "__main__":
    with open(file, 'r', encoding='utf-8') as f:
        # file = FileToTransform(name=file.name, file=f.read())
        file = f.read()
        file = {'files': file}

    with ThreadPoolExecutor(max_workers=stress) as executor:
        futures = [executor.submit(send_file, file, output, params) for _ in range(stress)]
        for future in futures:
            print(future.result())



from typing import List

from ..models import Pivot
from ..transformers.transformer import Transformer

import pandas as pd


class GephiTransformer(Transformer):
    def __init__(self):
        super(GephiTransformer, self).__init__()

    def transform(self, pivot_list: List[Pivot], graph_type="skeletton") -> str:
        keyword_to_authors = {}
        res = "Source;Target;edge_label"
        for pivot in pivot_list:
            for keyword in pivot.keywords.split(','):
                if keyword.strip() == "":
                    continue
                keyword_to_authors[keyword] = keyword_to_authors.get(keyword, set())
                keyword_to_authors[keyword].add(pivot.auteur)
        # pour chaque mot clé
        for keyword, author_list in keyword_to_authors.items():
            # pour chaque auteur de la liste des auteurs
            for index in range(len(list(author_list))):
                author = list(author_list)[index]
                if author.lower() != "unknown":
                    # ce code là sert pour créer des liens entre auteurs, les résultats ne sont pas probants
                    # # pour chaque autre auteur de la liste des auteurs
                    # for sub_index, author in enumerate(list(author_list)[index:]):
                    #     if sub_index < len(list(author_list)[index:]) - 2:
                    #         # on créé un lien
                    #         res += f"\n{author};{list(author_list)[sub_index + 1]};{keyword}"
                    # celui ci pour créer des liens auteur -> mot clé
                    res += f"\n{author.lower()};{keyword};aucun"

        return res
        # code pour le graphe temporel si on le fait un jour
        # pivot_list.sort(key=lambda x: x.date)
        # first_date = pivot_list[0].date
        # last_date = pivot_list[-1].date

        # return df.to_csv(sep=",", index=False)



from pathlib import Path
from typing import Optional

from . import pipeline
from .models import FileToTransform, TransformerOutput, Params, Output


def main(folder: Path | str, outputs: list[Output], params: Optional[Params] = None) -> None:
    if params is None:
        params = Params()

    if isinstance(folder, str):
        folder = Path(folder)
    elif not isinstance(folder, Path):
        raise ValueError(f"folder must be a Path or a string, not {type(folder)}")

    # parse all files
    files = []
    for file in folder.iterdir():
        if file.suffix.lower() != ".html":
            print(f"Skipping {file.name} (not an HTML file)")
            continue
        if file.is_file():
            with open(file, 'r', encoding='utf-8') as f:
                files.append(FileToTransform(name=file.name, file=f.read()))
    print(f"Processing {len(files)} files...")

    # process result
    results: list[TransformerOutput] = pipeline(files, outputs, params=params)

    # write results
    folder = folder / 'results'
    folder.mkdir(exist_ok=True)
    for result in results:
        print(result.filename)
        if isinstance(result.data, str):
            with open(folder / f'{result.filename}', 'w', encoding='utf-8') as f:
                f.write(result.data)
        else:
            with open(folder / f'{result.filename}', 'wb') as f:
                f.write(result.data)



if __name__ == '__main__':
    folder = Path('/home/marceau/Nextcloud/eurocollectes/15-18')
    outputs = ["json", "txm", "iramuteq", "csv", "excel", "stats", "processed_stats", "plots", "markdown"]
    # outputs = ["json", "stats", "processed_stats", "plots"]
    params = Params(
        minimal_support_kw=5,
        minimal_support_authors=2,
        minimal_support_journals=8,
        minimal_support_dates=3,
    )
    main(folder, outputs, params=params)
